#!/usr/bin/python

import string
import re
import sys
import os
import gzip
import math
from sequence import *
from alignment import *


def convert1LetterTo3Letter(code):
    if code=="A": return "ALA"
    elif code=="R": return "ARG"
    elif code=="N": return "ASN"
    elif code=="D": return "ASP"
    elif code=="C": return "CYS"
    elif code=="Q": return "GLN"
    elif code=="E": return "GLU"
    elif code=="G": return "GLY"
    elif code=="H": return "HIS"
    elif code=="I": return "ILE"
    elif code=="L": return "LEU"
    elif code=="K": return "LYS"
    elif code=="M": return "MET"
    elif code=="F": return "PHE"
    elif code=="P": return "PRO"
    elif code=="S": return "SER"
    elif code=="T": return "THR"
    elif code=="W": return "TRP"
    elif code=="Y": return "TYR"
    elif code=="V": return "VAL"
    else: return "UNK" #amino acid not identified


def convert3LetterTo1Letter(code):
    if code=="ALA": return "A"
    elif code=="ARG": return "R"
    elif code=="ASN": return "N"
    elif code=="ASP": return "D"
    elif code=="CYS": return "C"
    elif code=="GLN": return "Q"
    elif code=="GLU": return "E"
    elif code=="GLY": return "G"
    elif code=="HIS": return "H"
    elif code=="ILE": return "I"
    elif code=="LEU": return "L"
    elif code=="LYS": return "K"
    elif code=="MET": return "M"
    elif code=="PHE": return "F"
    elif code=="PRO": return "P"
    elif code=="SER": return "S"
    elif code=="THR": return "T"
    elif code=="TRP": return "W"
    elif code=="TYR": return "Y"
    elif code=="VAL": return "V"
    else: return "U"
#end function

compare_re = re.compile("[0-9]+[A-Z]$")
def compareResidueNumbers(res1,res2):
    chain1 = res1[-1]
    chain2 = res2[-1]
    res1 = res1[:-2]
    res2 = res2[:-2]

    res1 = int(res1)
    res2 = int(res2)
    if chain1==chain2:
        return cmp(res1,res2)
    elif chain1 < chain2:
        return -1
    else:
        return 1


#Gets all of the residues in the SSE number passed in through SSEnum
def getSegment(query,SSEnum):
    numSSEs = query.num_SSEs
    segment_residues = []

    res1,res2 = query.sse_ends[SSEnum]
    start_index = query.residue_indices[res1]
    end_index = query.residue_indices[res2]
    for res_idx in range(start_index,end_index+1):
        segment_residues.append(query.residues[res_idx])

    return segment_residues

#Constructs a HELIX line for a PDB file, where the helix starts at
#residue number start, ends at end, and has chain identifier chain
def getHelixLine(start,end,chain):
    line = "HELIX "+" "*13
    line = line+chain+" "
    line = line + "%4s" % start
    line = line + " "*6+chain+" "
    line = line + "%4s" % end
    line = line+"\n";
    return line

#Constructs a SHEET line for a PDB file
def getSheetLine(start,end,chain):
    line = "SHEET "+" "*15
    line = line+chain
    line = line + "%4s" % start + " "*6
    line = line + chain+"%4s" % end + "\n";
    return line

#Prints PDB files for a specified segment of residues in the query and target
#structures. The parameter segment_residues holds the residues for the query PDB file,
#and the parameter segment_residues2 holds the residues for the target PDB file.
#The old_correspondences and new_correspondences hold residue pairs that CBA
#considers to be aligned at this point.
def printSegmentPDBFile(tmp_subdir,query,target,querySSEnum,targetSSEnum,segment_residues,segment_residues2,old_correspondences,new_correspondences):
    querychain = query.chain
    if querychain != " " and querychain != "all": querychain = querychain.upper()
    else: querychain = " "

    targetchain = target.chain
    if targetchain != " " and targetchain != "all": targetchain = targetchain.upper()
    else: targetchain = " "    

    file = "/tmp/%s/%s_%i.pdb" % (tmp_subdir,query.domainname,querySSEnum)
    queryoutfile = open(file,"w")
    queryoutfile.write("HEADER    %s SSE number %i\n" % (query, querySSEnum))

    file = "/tmp/%s/%s_%i.pdb" % (tmp_subdir,target.domainname,targetSSEnum)
    targetoutfile = open(file,"w")
    targetoutfile.write("HEADER    %s SSE number %i\n" % (target, targetSSEnum)) 

    #include a header that gives the residues to lock down
    query_residues = ""
    target_residues = ""
    segment_members = dict([(x,1) for x in segment_residues])
    segment_members2 = dict([(x,1) for x in segment_residues2])
    num_correspondences = 0

    for queryresnum in old_correspondences:
        if not segment_members.has_key(queryresnum): continue
        targetresnum = old_correspondences[queryresnum]
        if not segment_members2.has_key(targetresnum): continue
        num_correspondences += 1
        query_residues  += "%4s%s" % (queryresnum[:-2],queryresnum[-2])
        target_residues += "%4s%s" % (targetresnum[:-2],targetresnum[-2])
    for queryresnum in new_correspondences:
        if not segment_members.has_key(queryresnum): continue
        targetresnum = new_correspondences[queryresnum]
        if not segment_members2.has_key(targetresnum): continue
        num_correspondences += 1
        query_residues  += "%4s%s" % (queryresnum[:-2],queryresnum[-2])
        target_residues += "%4s%s" % (targetresnum[:-2],targetresnum[-2])
    queryoutfile.write("REMARK     ")
    queryoutfile.write("%5i" % num_correspondences)
    queryoutfile.write(query_residues+"\n")
    targetoutfile.write("REMARK     ")
    targetoutfile.write("%5i" % num_correspondences)
    targetoutfile.write(target_residues+"\n")

    #print secondary structure information
    querystart,queryend = query.sse_ends[querySSEnum]
    targetstart,targetend = target.sse_ends[targetSSEnum]
    SSEtype = query.sstype[querystart]
    querychain = querystart[-1]
    targetchain = targetstart[-1]

    icode = querystart[-2]
    if icode==" ": icode=""
    querystart = querystart[:-2]+icode
    if queryend[-2]==" ":
        queryend = queryend[:-2]
    else: queryend = queryend[:-1]
    if targetstart[-2]==" ":
        targetstart = targetstart[:-2]
    else: targetstart = targetstart[:-1]
    if targetend[-2]==" ": targetend = targetend[:-2]
    else: targetend = targetend[:-1]
    if SSEtype == "H": #helix
        line = getHelixLine(querystart,queryend,querychain)
        queryoutfile.write(line)
        line = getHelixLine(targetstart,targetend,targetchain)
        targetoutfile.write(line)
    else:
        line = getSheetLine(querystart,queryend,querychain)
        queryoutfile.write(line)
        line = getSheetLine(targetstart,targetend,targetchain)        
        targetoutfile.write(line)
        
    count = 0
    for resnum in segment_residues:
        i = query.residue_indices[resnum]
        chain = resnum[-1]
        if chain=="-": chain = " "
        atoms = query.residue_atomnames[i]
        for atom in atoms:
            if not (atom=="CA"  or atom=="CB" or atom=="C" or atom=="N" or atom=="O"): continue
            line = "ATOM  " #okay
            line += "%5s " % count
            line += "%-4s %s %s" % (atom,convert1LetterTo3Letter(query.aas[i]),chain) #okay
            line += "%4s%s" % (query.residues[i][:-2],query.residues[i][-2])#27 
            line += "   %8.3f%8.3f%8.3f\n" % (query.coordinates[i][atom][0],query.coordinates[i][atom][1],query.coordinates[i][atom][2])                                
            queryoutfile.write(line)
            count += 1

    count = 0
    for resnum in segment_residues2:
        i = target.residue_indices[resnum]
        chain = resnum[-1]
        if chain=="-": chain = " "
        atoms = target.residue_atomnames[i]
        for atom in atoms:
            if not (atom=="CA"  or atom=="CB" or atom=="C" or atom=="N" or atom=="O"): continue            
            line = "ATOM  " #okay
            line += "%5s " % count
            line += "%-4s %s %s" % (atom,convert1LetterTo3Letter(target.aas[i]),chain) #okay
            line += "%4s%s" % (target.residues[i][:-2],target.residues[i][-2])#27
            line += "   %8.3f%8.3f%8.3f\n" % (target.coordinates[i][atom][0],target.coordinates[i][atom][1],target.coordinates[i][atom][2])                
            targetoutfile.write(line)
            count += 1
    queryoutfile.write("END\n")
    targetoutfile.write("END\n")
    queryoutfile.close()
    targetoutfile.close()


begin_alignment_match = re.compile("^Num Core Aligned")
query_transformation_re=re.compile("Transformation Matrix for query")
target_transformation_re=re.compile("Transformation Matrix for target")

#Reads the LOCK 2 alignment for the superposition of a query and target SSE
def readLockResults(alignmentscores,gapscores,directory):
    try:
        alignment = open("%s/query_target.out" % directory,"r")
    except:
        printError("Error: Can't open LOCK 2 SSE realignment results: %s/query_target.out" % directory)
    
    line = alignment.readline()
    core_aligned = 0
    while line:
        if begin_alignment_match.match(line):
            core_aligned = 1;
            dummy, num_aligned = re.split(" = ",line)
            num_aligned = string.atoi(num_aligned)
            break
        line = alignment.readline()
    
    query_resnums = []; target_resnums = []
    if not core_aligned:
        alignment.close()
        return query_resnums,target_resnums,identity(3,Int),zeros([3],Int)

    for i in range(num_aligned):
        line = alignment.readline()
        fields = re.split("\t",line)
        resnum1 = fields[0][:-1].strip() #get rid of amino acid
        resnum2 = fields[1][:-1].strip()
        score = float(fields[3])
        gap = float(fields[-1])
        query_resnums.append(resnum1); target_resnums.append(resnum2)
        alignmentscores.setdefault(resnum1,[]).append((resnum2,score))
        gapscores.setdefault(resnum1,[]).append((resnum2,gap))

    while line:
        if query_transformation_re.match(line):
            query_rotation = zeros([3,3],Float)
            for i in range(3):
                line=alignment.readline()
                fields = map(float,line.split()[:-1])
                for j in range(3):
                    query_rotation[i][j] = fields[j]
            line = alignment.readline()
            query_translation = array(map(float,line.split()[:-1]))
        elif target_transformation_re.match(line):
            target_rotation = zeros([3,3],Float)
            for i in range(3):
                line=alignment.readline()
                fields = map(float,line.split()[:-1])
                for j in range(3):
                    target_rotation[i][j] = fields[j]
            line=alignment.readline()
            target_translation = array(map(float,line.split()[:-1]))
        line=alignment.readline()

    #Make the query stationary
    if not len(query_resnums)==0:
        query_rotation_inverse = inverse(query_rotation)
        new_target_translation = dot(target_translation-query_translation,query_rotation_inverse)
        new_target_rotation = dot(target_rotation,query_rotation_inverse)
    else:
        new_target_rotation = identity(3,Float)
        new_target_translation = zeros([3],Float)
    return(query_resnums,target_resnums,new_target_rotation,new_target_translation)


#Prints the alignment of the query vs target structures after all SSEs have been
#realigned. The parameter new_aligned_residues holds the residues that are now
#aligned in the new pairwise superposition.
def printAlignment(newalignmentdir,query,target,alignment,alignmentscores,gapscores,new_aligned_residues):
    try:
        outfile=open(os.path.join(newalignmentdir, "%s_%s.out" % (query.domainname,target.domainname)),"w")
    except:
        printError("Error: Couldn't open fixed alignment file for writing")

    outfile.write("Query: %s/%s%s\n" % (query.directory,query.domainname,query.extension));
    outfile.write("Target: %s/%s%s\n" % (target.directory,target.domainname,target.extension))
    if not new_aligned_residues.has_key(query.domainname):
        mykeys = new_aligned_residues.keys()
    query_residues = new_aligned_residues[query.domainname][target.domainname].keys()
    query_residues.sort(compareResidueNumbers)
    num_aligned = len(query_residues)
    outfile.write("Num Core Aligned = %i\n" % num_aligned)
    for i in range(num_aligned):
        target_residue = new_aligned_residues[query.domainname][target.domainname][query_residues[i]]
        outfile.write("%s %s\t%s %s\t-1\t%f\t%f\n" % (query_residues[i],query.aas[query.residue_indices[query_residues[i]]],new_aligned_residues[query.domainname][target.domainname][query_residues[i]],target.aas[target.residue_indices[new_aligned_residues[query.domainname][target.domainname][query_residues[i]]]],alignmentscores[query_residues[i]],gapscores[query_residues[i]]))
    outfile.writelines(alignment.finallines)
    outfile.close()


    
#Looks at the current multiple structure alignment to determine which
#residues are aligned in each pairwise alignment. Determines which correspondences
#are retained from the original LOCK 2 pairwise alignments and which ones have been
#added or changed by CBA so far.
def parseMSAInfo(nodes,clusters,domains,alignments,registration_changes,same_registration):
    domain_names = [domain.domainname for domain in domains]
    num_clusters = len(clusters)
    for i in range(num_clusters):
        cluster = clusters[i]
        for idx in range(len(cluster)):
            node = nodes.getNodeByLabel(cluster[idx])
            queryname = domain_names[node.domain_idx]
            query_resnum = node.resnum
            for query_idx in range(len(domains)):
                if domains[query_idx].domainname == queryname:
                    query = domains[query_idx]
                    break
            else:
                continue

            query_SSEnum = query.ssNumber[query_resnum]
            for idx2 in range(idx,len(cluster)):
                node2 = nodes.getNodeByLabel(cluster[idx2])
                targetname = domain_names[node2.domain_idx]
                target_resnum = node2.resnum
                for target_idx in range(len(domains)):
                    if domains[target_idx].domainname==targetname:
                        target = domains[target_idx]
                        break
                else:
                    continue
                target_SSEnum = target.ssNumber[target_resnum]

                #multiple structure alignment says that query_resnum and target_resnum are aligned
                if not alignments.alignments[query_idx][target_idx].alignedResidues.has_key(query_resnum):
                    #this fills in a gap in the pairwise alignment
                    registration_changes[query.domainname][target.domainname][query_resnum] = target_resnum
                    registration_changes[target.domainname][query.domainname][target_resnum]=query_resnum
                elif alignments.alignments[query_idx][target_idx].alignedResidues[query_resnum]!=target_resnum:
                    #this is a registration change
                    registration_changes[query.domainname][target.domainname][query_resnum]=target_resnum
                    registration_changes[target.domainname][query.domainname][target_resnum]=query_resnum
                else: #registration stayed the same
                    same_registration[query.domainname][target.domainname][query_resnum]=target_resnum
                    same_registration[target.domainname][query.domainname][target_resnum]=query_resnum


#Sometimes the same residue is aligned twice. This function finds those
#cases and retains the correspondence with the highest score.
def resolveAlignmentConflicts(query,target,aligned_residues,alignmentscores,gapscores):
    new_query_residues = []
    new_target_residues = []
    for res1 in aligned_residues.keys():
        choices = alignmentscores[res1]
        if(len(choices)) == 1:
            new_query_residues.append(res1); new_target_residues.append(alignmentscores[res1][0][0])
            alignmentscores[res1] = alignmentscores[res1][0][1]
            gapscores[res1] = gapscores[res1][0][1]
            continue
        score1 = alignmentscores[res1][0][1] + gapscores[res1][1][1]
        score2 = alignmentscores[res1][1][1] + gapscores[res1][0][1]
        if score1 > score2:
            new_target_residues.append(alignmentscores[res1][0][0])
            alignmentscores[res1] = alignmentscores[res1][0][1]
            gapscores[res1] = gapscores[res1][0][1]            
        else:
            new_target_residues.append(alignmentscores[res1][1][0])
            alignmentscores[res1] = alignmentscores[res1][1][1]
            gapscores[res1] = gapscores[res1][1][1]            
        new_query_residues.append(res1)
    return new_query_residues,new_target_residues

def getGapScore(domain,resnum,openSSE):
    multiplier = 6
    SSE_open_multiplier = 3
    if domain.sstype[resnum]=='U':
        if openSSE:
            return SSE_open_multiplier*multiplier
        else:
            return multiplier*2
    else:
        if openSSE: return SSE_open_multiplier*multiplier
        else: return multiplier
    
#Returns the distance between point1 and point2 (typically two beta carbon positions)
def vectorDistance(point1,point2):
    return (math.sqrt((point1[0]-point2[0])*(point1[0]-point2[0])+(point1[1]-point2[1])*(point1[1]-point2[1])+(point1[2]-point2[2])*(point1[2]-point2[2])))


#Determines the angle between two vectors. The first vector is defined by point1a and
#point1b. The second is defined by point2a and point2b. Typically point1a and point2a are
#alpha carbon positions of two different residues, and point1b and point2b are the beta
#carbon positions of the same residues. This is the angle used by LOCK 2 in the residue
#alignment phase.
def vectorAngle(point1a,point1b,point2a,point2b):
    u = [0,0,0]; v=[0,0,0]
    for i in range(3):
        u[i] = point1a[i]-point1b[i]
        v[i] = point2a[i]-point2b[i]

    len1=0;len2=0;dotproduct=0
    epsilon=0.0001
    for i in range(3):
        len1 += u[i]*u[i]; len2 += v[i]*v[i]
    len1 = math.sqrt(len1); len2=math.sqrt(len2)
    for i in range(3):
        u[i] /= float(len1); v[i] /= float(len2)
        dotproduct += u[i]*v[i]
    if dotproduct<= 1+epsilon and dotproduct >= 1-epsilon:
        angle = 0
    elif dotproduct<=-1+epsilon and dotproduct >= -1+epsilon:
        angle=2*3.14159;
    else:
        angle = math.acos(dotproduct)
    return angle

#Computes the score for aligning two residues according to the scoring
#function used by LOCK 2.
def getAlignmentScore(query,target,query_res,target_res,target_res_coords):
    query_idx = query.residue_indices[query_res]
    target_idx = target.residue_indices[target_res]
    tdist = vectorDistance(query.coordinates[query_idx]["CB"],target_res_coords["CB"])
    query_ssetype = query.sstype[query_res]
    target_ssetype = target.sstype[target_res]
    if(query_ssetype==target_ssetype):
        dist_cutoff = 6.0
        angle_cutoff = 0.6168 #pi/4*pi/4
    else:
        dist_cutoff = 3.0
        angle_cutoff = 2.467401 #pi/2*pi/2
    if tdist <= dist_cutoff:
        dist_score = 10.0/(1+tdist*tdist/16.0)-5.0
    else:
        dist_score = -10
    angle = vectorAngle(query.coordinates[query_idx]["CA"],query.coordinates[query_idx]["CB"],target_res_coords["CA"],target_res_coords["CB"])
    angle_score = 30.0/(1+angle*angle/angle_cutoff)-15.0
    return dist_score+angle_score

def dot(array1, array2):
    return innerproduct(array1, swapaxes(array2, -1, -2))

def transformCoordinate(domain,rotation,translation,resnum):
    new_coordinate = {}
    resindex = domain.residue_indices[resnum]
    atomnames = domain.residue_atomnames[resindex]
    for atom in atomnames:
        new_coordinate[atom] = array(dot(domain.coordinates[resindex][atom],rotation)+translation)
    return new_coordinate
    

#Realigns each aligned SSE in each pairwise alignment. The parameter msafile is not needed. (I
#just haven't gone back and changed the code that calls fixSSEs, so it's still there.)
def fixSSEs(domains,alignments,clusters,nodes,alignmentdir,msafile,newalignmentdir,lockprogram,tmp_subdir,gapOpenPenalty,gapExtendPenalty,distanceThreshold):

    #Temporary files will be written to /tmp. Make sure a subdirectory is chosen
    #that doesn't already exist.
    tmp_subdir = int(tmp_subdir)
    while os.path.exists("/tmp/%i" % tmp_subdir):
        tmp_subdir += 1

    try:
        os.mkdir("/tmp/%i" % tmp_subdir)
    except:
        printError("Error: Couldn't create directory /tmp/%i" % tmp_subdir)

    if not os.path.exists(newalignmentdir):
        try:
            os.mkdir(newalignmentdir)
        except:
            printError("Error: Couldn't create fixed alignments directory "+newalignmentdir)

    
    #Initialize registration changes. Will hold information about registrations in
    #the multiple alignment compared to the pairwise alignments.
    registration_changes = {}
    same_registration = {}
    for domain in domains:
        registration_changes[domain.domainname] = {}
        same_registration[domain.domainname] = {}
        for domain2 in domains:
            registration_changes[domain.domainname][domain2.domainname] = {}
            same_registration[domain.domainname][domain2.domainname] = {}
    

    #Will hold residues that are aligned after fixing SSEs
    new_aligned_residues = {}
    for domain in domains:
        new_aligned_residues[domain.domainname] = {}
        for domain2 in domains:
            new_aligned_residues[domain.domainname][domain2.domainname] = {}


    
    parseMSAInfo(nodes,clusters,domains,alignments,registration_changes,same_registration)


    for i in range(len(domains)):
        query = domains[i]
        query_SSEs = query.sse_ends

        for j in range(i,len(domains)):
            target = domains[j]
            transformed_coordinates = None #don't transform unless necessary

            new_residues = []
            alignmentscores = {}
            gapscores = {}
            all_alignmentscores = {}
            all_gapscores = {}
            for SSEnum in range(len(query_SSEs)):                
                if not alignments.alignments[i][j].alignedSSEs.has_key(query_SSEs[SSEnum][0]): continue
                start,end = query_SSEs[SSEnum]
                segment_residues = getSegment(query,SSEnum)
                target_start = alignments.alignments[i][j].alignedSSEs[start]
                target_SSEnum = target.ssNumber[target_start]
                segment_residues2 = getSegment(target,target_SSEnum)
                segment_members = dict([(x,1) for x in segment_residues])
                segment_members2 = dict([(x,1) for x in segment_residues2])
                

                #At this point, we have an aligned secondary structure element that
                #needs to be realigned. The query residues in the SSE are in segment_residues,
                #and the target residues in the SSE are in segment_residues2.

                #look for segment residues that are aligned to loop residues or residues from
                #a different SSE and add them to the segments
                for res in segment_residues:
                    if alignments.alignments[i][j].alignedResidues.has_key(res):
                        aligned_res = alignments.alignments[i][j].alignedResidues[res]
                        if not segment_members2.has_key(aligned_res):
                            segment_residues2.append(aligned_res)
                            segment_members2[aligned_res]=1
                for res in segment_residues2:
                    if alignments.alignments[j][i].alignedResidues.has_key(res):
                        aligned_res = alignments.alignments[j][i].alignedResidues[res]
                        if not segment_members.has_key(aligned_res):
                            segment_residues.append(aligned_res)
                            segment_members[aligned_res]=1
                segment_residues.sort(compareResidueNumbers)
                segment_residues2.sort(compareResidueNumbers)

                same = {} #Correspondences in this region of the multiple alignment that are the same as in the pairwise alignment
                changes = {} #Correspondences in this region of the multiple alignment that differ from pairwise alignment
                for res in registration_changes[query.domainname][target.domainname].keys():
                    if segment_members.has_key(res):
                        changes[res] = registration_changes[query.domainname][target.domainname][res]
                for res in segment_residues:
                    new_residues.append(res)
                    if alignments.alignments[i][j].alignedResidues.has_key(res): #if residue was aligned in pairwise
                        old_residue = alignments.alignments[i][j].alignedResidues[res]
                        new_residue = -1
                        if same_registration[query.domainname][target.domainname].has_key(res):
                            new_residue = same_registration[query.domainname][target.domainname][res]
                            if new_residue==old_residue and new_residue!=-1: #registration is the same in MSA and pairwise
                                same[res] = same_registration[query.domainname][target.domainname][res]
                    

                querystart = start; queryend = end
                targetstart,targetend = target.sse_ends[target_SSEnum]
 
                #Print out PDB files for the query and target segments. Information will
                #be included in a REMARK line that LOCK 2 will use to figure out which
                #residues to use to create an initial superposition of the target onto
                #the query
                printSegmentPDBFile(tmp_subdir,query,target,SSEnum,target_SSEnum,segment_residues,segment_residues2,same,changes)


                #Now run the LOCK 2 alignment. The -sse argument tells LOCK 2 to skip the SSE
                #alignment phase and instead superimpose the target onto the query using the
                #residue correspondences in the REMARK lines of the query and target pdb files.
                command = "%s -q /tmp/%s/%s_%i.pdb -t /tmp/%s/%s_%i.pdb -open %s -extend %s -d %f -sse -o /tmp/%s> /dev/null" % (lockprogram,tmp_subdir,query.domainname,SSEnum,tmp_subdir,target.domainname,target_SSEnum,gapOpenPenalty,gapExtendPenalty,distanceThreshold,tmp_subdir)
                os.system(command)


                query_residues,target_residues,new_rotation,new_translation = readLockResults(alignmentscores,gapscores,"/tmp/%s" % tmp_subdir)

                debug = 0

                #Add the aligned residues to the dictionary new_aligned_residues[query][target],
                #keeping in mind that the same residue may be aligned in two different SSE
                #realignments.
                for idx in range(len(query_residues)): 
                    query_resnum = query_residues[idx]; target_resnum = target_residues[idx]
                    if not new_aligned_residues[query.domainname][target.domainname].has_key(query_resnum):
                        new_aligned_residues[query.domainname][target.domainname][query_resnum] = []
                        all_alignmentscores[query_resnum] = []
                        all_gapscores[query_resnum] = []
                    new_aligned_residues[query.domainname][target.domainname][query_resnum].append(target_resnum)
                    for idx2 in range(len(alignmentscores[query_resnum])):
                        all_alignmentscores[query_resnum].append(alignmentscores[query_resnum][idx2])
                        all_gapscores[query_resnum].append(gapscores[query_resnum][idx2])


                #This loop is necessary to make sure that residues included in the segment
                #that are not in the SSE itself and that do not align in the realignment
                #still retain their correspondences from the pairwise alignments. This helps
                #insure that overzealous inclusion of residues in the segments to be realigned
                #does not do any harm if the residues are not aligned in the realignment. I think.
                for idx in range(len(segment_residues)):
                    if not segment_residues[idx] in query_residues: #if not aligned in new alignment, but aligned originally
                        query_res = segment_residues[idx]
                        if alignments.alignments[i][j].alignedResidues.has_key(segment_residues[idx]):
                            target_res = alignments.alignments[i][j].alignedResidues[segment_residues[idx]]
                            if not new_aligned_residues[query.domainname][target.domainname].has_key(query_res):
                                new_aligned_residues[query.domainname][target.domainname][query_res] = []
                                all_alignmentscores[query_res] = []
                                all_gapscores[query_res] = []
                            for query_res2,aligned_residues in new_aligned_residues[query.domainname][target.domainname].items():
                                aligned_res = target_res
                                if len(new_aligned_residues[query.domainname][target.domainname][query_res2])==0: continue
                                if aligned_res==new_aligned_residues[query.domainname][target.domainname][query_res2][0]: #aligned_res already aligned to query_res, so can't just add res vs query_res2
                                    aligned_res_coords = transformCoordinate(target,new_rotation,new_translation,aligned_res)
                                    new_score = getAlignmentScore(query,target,query_res2,aligned_res,aligned_res_coords)
                                    old_score = getAlignmentScore(query,target,query_res,aligned_res,aligned_res_coords)
                                    openSSE_new = 0
                                    openSSE_old = 0
                                    query_residx = query.residue_indices[query_res]
                                    if query_residx-1 >= 0:
                                        if query.sstype[query.residues[query_residx-1]]!='U' and not query.isSSEEnd(query_residx-1)==1:
                                            openSSE_new = 1
                                    query_residx = query.residue_indices[query_res2]
                                    if query_residx-1>=0:
                                        if query.sstype[query.residues[query_residx-1]]!='U' and not query.isSSEEnd(query_residx-1)==1:
                                            openSSE_old = 1
                                    gap_score_new = getGapScore(query,query_res,openSSE_new)
                                    gap_score_old = getGapScore(query,query_res2,openSSE_old)
                                    new_score -= gap_score_new
                                    old_score -= gap_score_old

                                    if old_score >= new_score: #will be using the old correspondence
                                        new_aligned_residues[query.domainname][target.domainname][query_res].append(aligned_res)
                                        all_alignmentscores[query_res].append((aligned_res,alignments.alignments[i][j].alignmentscores[query_res]))
                                        all_gapscores[query_res].append((aligned_res,alignments.alignments[i][j].gapscores[query_res]))
                                        del new_aligned_residues[query.domainname][target.domainname][query_res2]
                                        del all_alignmentscores[query_res2]
                                        del all_gapscores[query_res2]
                                    else:
                                        del new_aligned_residues[query.domainname][target.domainname][query_res]
                                        del all_alignmentscores[query_res]
                                        del all_gapscores[query_res]
                                    break

                            else:        
                                new_aligned_residues[query.domainname][target.domainname][query_res].append(target_res)
                                all_alignmentscores[query_res].append((target_res,alignments.alignments[i][j].alignmentscores[query_res]))
                                all_gapscores[query_res].append((target_res,alignments.alignments[i][j].gapscores[query_res]))

            query_residues,target_residues = resolveAlignmentConflicts(query,target,new_aligned_residues[query.domainname][target.domainname],all_alignmentscores,all_gapscores)
            new_aligned_residues[query.domainname][target.domainname] = {}
            for idx in range(len(query_residues)):
                new_aligned_residues[query.domainname][target.domainname][query_residues[idx]] = target_residues[idx]

            debug = 0
            if debug: print "adding in old residues"
            for res in query.residue_indices.keys():
                if debug: print "checking ",res
                if not res in new_residues:
                    if alignments.alignments[i][j].alignedResidues.has_key(res):
                        if debug: print "   not in list of new aligned query residues and was aligned to %s" % alignments.alignments[i][j].alignedResidues[res]
                        aligned_res = alignments.alignments[i][j].alignedResidues[res]
#                        print "   aligned_res=",aligned_res
                        for query_res in new_aligned_residues[query.domainname][target.domainname].keys():
#                            print "    checking to see if query residue %s is aligned to target residue %s" % (query_res,aligned_res),new_aligned_residues[query.domainname][target.domainname][query_res]
                            if new_aligned_residues[query.domainname][target.domainname][query_res]==aligned_res and not (query_res in segment_members and aligned_res in segment_members2):

                                if debug: print "conflict between query residues %s and %s with target residue %s" % (query_res,res,aligned_res)
                                aligned_res_coords = transformCoordinate(target,new_rotation,new_translation,aligned_res)
                                new_score = getAlignmentScore(query,target,query_res,aligned_res,aligned_res_coords)
                                old_score = alignments.alignments[i][j].alignmentscores[res]

                                if debug: print "old score, new score:",old_score,new_score
                                openSSE = 0
                                target_residx = target.residue_indices[aligned_res]
                                if target_residx-1 >= 0:
                                    if target.sstype[target.residues[target_residx-1]]!='U' and not target.isSSEEnd(target_residx-1)==1:
                                        openSSE = 1
                                
                                gap_score_old = getGapScore(query,query_res,openSSE)
                                gap_score = getGapScore(query,res,openSSE)
                                new_score -= gap_score
                                old_score -= gap_score_old
                                if debug: print "gap scores:",gap_score_old,gap_score
                                if debug: print "old score, new score:",old_score,new_score
                                if old_score >= new_score: #will be using the old correspondence
                                    if debug: print "deleting alignment between %s and %s" % (query_res,aligned_res)
                                    if debug: print "adding %s and %s" % (res,aligned_res)
                                    new_aligned_residues[query.domainname][target.domainname][res] = aligned_res
                                    all_alignmentscores[res] = alignments.alignments[i][j].alignmentscores[res]
                                    all_gapscores[res] = alignments.alignments[i][j].gapscores[res]
                                    del new_aligned_residues[query.domainname][target.domainname][query_res]
                                    del all_alignmentscores[query_res]
                                    del all_gapscores[query_res]


                                break
                        else: #no conflict
                            if debug: print "no conflict: adding %s and %s" % (res,aligned_res)                            
                            new_aligned_residues[query.domainname][target.domainname][res] = aligned_res
                            all_alignmentscores[res] = alignments.alignments[i][j].alignmentscores[res]
                            all_gapscores[res] = alignments.alignments[i][j].gapscores[res]

            if debug: print "new_aligned_residues:"
            for query_res,target_res in new_aligned_residues[query.domainname][target.domainname].items():
                if debug: print query_res,target_res

            printAlignment(newalignmentdir,query,target,alignments.alignments[i][j],all_alignmentscores,all_gapscores,new_aligned_residues)
            
    try:
        os.system("/bin/rm -rf /tmp/%i" % (tmp_subdir))
    except:
        printError("Error: Couldn't remove directory /tmp/%i" % tmp_subdir)


if 1:
    try:
        import psyco
        psyco.full()
    except:
        pass
